#!/usr/bin/env python

import os
import sys
import logging
import gzip
from .helper import return_filehandle


def count_gff_features(gff):
    counts = {}
    fh = return_filehandle(gff)  # use magic filehandle method
    if not fh:
        logger.error('could not open {}'.format(gff))
        sys.exit(1)
    with fh as fopen:
        for line in fopen:
            if not line or line.isspace() or line.startswith('#'):  # skip comments
                continue
            line = line.rstrip()
            f = line.split('\t')  # get fields
            if f[2] not in counts:  # feature type
                counts[f[2]] = 1
                continue
            counts[f[2]] += 1
    fh.close()
    return counts


if __name__ == '__main__':
    print('import me to use stast for genome_main and gene_models_main.' +
          'This should be in a class probably... but its w/e for now')
    sys.exit(1)



#!/usr/bin/env python

import os
import sys
import gzip
import errno
import logging
import subprocess


def check_subprocess_dependencies():
    '''attempts to run samtools, tabix and bgzip'''
    subprocess.check_call('which bgzip', shell=True)  # check in env
    subprocess.check_call('which tabix', shell=True)  # check in env
    subprocess.check_call('which samtools', shell=True)  # check in env
    subprocess.check_call('which gffread', shell=True)  # check in env
    return True


def check_busco_dependencies():
    '''attempts to run samtools, tabix and bgzip'''
    subprocess.check_call('which run_BUSCO.py', shell=True)  # check in env
    subprocess.check_call('which augustus', shell=True)  # check in env
    return True


def setup_logging(log_file, log_level):
    '''Return logger based on log_file and log_level'''
    log_level = getattr(logging, log_level.upper(), logging.INFO)
    msg_format = '%(asctime)s|%(name)s|[%(levelname)s]: %(message)s'
    logging.basicConfig(format=msg_format, datefmt='%m-%d %H:%M',
                        level=log_level)
    log_handler = logging.FileHandler(log_file, mode='w')
    formatter = logging.Formatter(msg_format)
    log_handler.setFormatter(formatter)
    logger = logging.getLogger('detect_incongruencies')
    logger.addHandler(log_handler)
    return logger


def return_filehandle(open_me):
    '''return file handle for gz compressed or text file'''
    magic_dict = {  # headers for compression
        b'\x1f\x8b\x08': 'gz',
        #                 '\x42\x5a\x68': 'bz2',
        #                 '\x50\x4b\x03\x04': 'zip'
    }
    max_bytes = max(len(t) for t in magic_dict)
    with open(open_me, 'rb') as f:
        s = f.read(max_bytes)
    for m in magic_dict:
        if s.startswith(m):  # check file header for match with m
            t = magic_dict[m]
            if t == 'gz':
                return gzip.open(open_me, 'rt')
  #          elif t == 'bz2':
  #              return bz2.open(open_me)
  #          elif t == 'zip':
  #              return zipfile.open(open_me)
    return open(open_me)


def check_file(f):
    '''check for file using os.path.isfile'''
    try:
        os.path.isfile(f)
    except OSError:
        raise
    return os.path.isfile(f)


def create_directories(dirpath):
    '''make directory path'''
    try:
        os.makedirs(dirpath)
    except OSError as e:
        if e.errno != errno.EEXIST:  # ignore if error is exists else raise
            raise


def change_directories(dirpath):
    '''cd to path, does not modify calling process CWD'''
    try:
        os.chdir(dirpath)
    except OSError as e:
        raise


if __name__ == '__main__':
    print('import me to use check_file and return_filehandle.' +
          'This should be in a class probably')
    sys.exit(1)



#!/usr/bin/env python

import os
import sys
import logging
import re
import subprocess
import gzip
import click
from ruamel.yaml import YAML
from .helper import check_file, return_filehandle, create_directories,\
    change_directories, check_busco_dependencies, setup_logging


def preprocess_input(target):
    '''Creates a temporary file for BUSCO to act on and sets temp and output'''
    name = os.path.basename(target)
    target_dir = os.path.dirname(target)
    busco_dir = '{}/busco'.format(target_dir)
    create_directories(busco_dir)  # make runtime dir
    change_directories(busco_dir)  # go to runtime dir cause busco cannot path
    temp_file = './{}_buscoready.fa'.format(name)
    temp_file_handle = open(temp_file, 'w')
    temp_target = return_filehandle(target)
    temp_file_handle.writelines(temp_target)
    temp_file_handle.close()
    return temp_file


def run_busco(busco_target, output, temp_dir, threads, mode, lineage):
    '''takes busco_target and performs BUSCO using lineage and mode'''
    cmd = '''run_BUSCO.py -f -m {} -c {} -i {} -l {} -t {} -o {}'''.format(
        mode, threads, busco_target, lineage, temp_dir, output)
    subprocess.check_call(cmd, shell=True)
    return output


@click.command()
@click.option('--target', type=str,
              help='''TARGETS can be files or directories or both''')
@click.option('--lineage', type=str,
              help='''BUSCO lineage to compare target against''')
@click.option('--mode', type=str,
              help='''BUSCO mode (genome, protein, transcript)''')
@click.option('--threads', type=int, default=4,
              help='''Threads for BUSCO''')
@click.option(
    '--log_file',
    metavar='<FILE>',
    default='./normalizer_index.log',
    help='''File to write log to. (default:./normalizer_index.log)''')
@click.option(
    '--log_level',
    metavar='<LOGLEVEL>',
    default='INFO',
    help='''Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)''')
def cli(target, lineage, mode, threads, log_file, log_level):
    '''Determines what type of index to apply to input target'''
    check_busco_dependencies()
    logger = setup_logging(log_file, log_level)
    if not (target and lineage):
        logger.error('--target, --lineage and arguments required')
        sys.exit(1)
    target = os.path.abspath(target)
    lineage = os.path.abspath(lineage)
    file_types = ['fna', 'fasta', 'fa', 'faa', 'ADD MORE']
    canonical_types = {'genome_main': 'genome', 'mrna': 'transcriptome',
                       'protein': 'proteins',
                       'protein_primaryTranscript': 'proteins'}
    fasta = ['fna', 'fasta', 'fa', 'faa']
    file_attributes = target.split('.')
    output = False
    if len(file_attributes) < 2:
        error_message = '''Target {} does not have a type or attributes.  File must end in either gz, bgz, fasta, fna, fa.'''.format(
            target)
        logger.error(error_message)
        sys.exit(1)
    if not (file_attributes[-2].lower() in canonical_types or file_attributes[-3].lower()
            in canonical_types) and not mode:
        logger.info(
            'file {} was not recognized as a canonical type and mode cannot be assigned please specify --mode for BUSCO to run anyway'.format(target))
        sys.exit(1)
    if file_attributes[-2].lower() in canonical_types and not mode:
        mode = canonical_types[file_attributes[-2].lower()]
    elif file_attributes[-3].lower() in canonical_types and not mode:
        mode = canonical_types[file_attributes[-3].lower()]
    if not mode:
        logger.info(
            'Mode was not assigned and could not be autodetected, please specify BUSCO mode with --mode')
        sys.exit(1)
    logger.info('Setting up output and making temp input...')
    busco_me = preprocess_input(target)
    name = os.path.basename(target)
    output = '{}_busco'.format(name)
    temp_dir = '{}_temp'.format(name)
    logger.info(
        'Runnig BUSCO on {} with mode {} and lineage {}...'.format(
            target, mode, lineage))
    output = run_busco(busco_me, output, temp_dir, threads, mode, lineage)
    logger.info('BUSCO done output: {}'.format(output))



#!/usr/bin/env python

import os
import sys
import logging
import re
import subprocess
import gzip
import click
from ruamel.yaml import YAML, RoundTripDumper
from sequencetools.helpers import sequence_helpers
from ..helper import check_file, return_filehandle, create_directories,\
    check_subprocess_dependencies, setup_logging


def primary_transcript_check(peptides, logger):
    '''Decides if protein file form run_gffread is primary of not

       if the file is not primary, it contains isoforms, the longest

       is selected for each gene and a protein_primaryTranscript file

       is produced.
    '''
    seq_handle = open(peptides, 'rt')
    longest = {}
    count = 0
    primary = '{}.protein_primaryTranscript.faa'.format(
        '.'.join(peptides.split('.')[:-2]))
    for record in sequence_helpers.get_seqio_fasta_record(seq_handle):
        count += 1
        my_record = '>{}\n{}'.format(record.id, record.seq)
        my_obj = {'record': my_record, 'length': len(record.seq)}
        gene_id = '.'.join(record.id.split('.')[:-1])
        if gene_id not in longest:
            longest[gene_id] = my_obj
        else:
            if len(record.seq) > longest[gene_id]['length']:
                longest[gene_id] = my_obj
    if len(longest) == count:
        logger.info('Protein file is already primary transcripts')
        return  # protein file is already primary
    primary_handle = open(primary, 'w')
    for r in longest:
        primary_handle.write(longest[r]['record'] + '\n')
    primary_handle.close()
    return primary


def run_gffread(gff, reference, logger):
    '''Reads gff3 file and writes mRNA, CDS and Peptides'''
    gff_dir = os.path.dirname(gff)
    gff_attributes = os.path.basename(gff).split('.')
    mrna = '{}/{}.mrna.fna'.format(gff_dir, '.'.join(gff_attributes[:5]))
    cds = '{}/{}.cds.fna'.format(gff_dir, '.'.join(gff_attributes[:5]))
    pep = '{}/{}.protein.faa'.format(gff_dir, '.'.join(gff_attributes[:5]))
    cmd = 'gffread {} -g {} -w {} -x {} -y {} -W'.format(gff, reference, mrna,
                                                         cds, pep)
    subprocess.check_call(cmd, shell=True)
    primary_transcript_check(pep, logger)


@click.command()
@click.option(
    '--target',
    type=str,
    help='''Formatted file from normalizer prefix MUST BE UNCOMPRESSED''')
@click.option(
    '--reference',
    type=str,
    help='''FASTA file of sequences to read gff intervals normalizer prefix is a good start here too MUST BE UNCOMPRESSED''')
@click.option(
    '--log_file',
    metavar='<FILE>',
    default='./normalizer_index.log',
    help='''File to write log to. (default:./normalizer_index.log)''')
@click.option(
    '--log_level',
    metavar='<LOGLEVEL>',
    default='INFO',
    help='''Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)''')
def cli(target, reference, log_file, log_level):
    '''Determines what typae of index to apply to input target'''
    logger = setup_logging(log_file, log_level)
    check_subprocess_dependencies()
    if not (target and reference):
        logger.error('--target and --reference arguments are required')
        sys.exit(1)
    target = os.path.abspath(target)  # get full path
    target_attributes = os.path.basename(target).split('.')
    if os.path.basename(reference).split('.')[-1] == 'gz':
        logger.error('GFFREAD cannot process compressed fasta as reference')
        sys.exit(1)
    if len(target_attributes) < 7:
        logger.error(
            'Target file {} is not delimited correctly'.format(target))
        sys.exit(1)
    run_gffread(target, reference, logger)  # write readme template



#!/usr/bin/env python

import os
import sys
import re
import click
import logging
import subprocess
from ruamel.yaml import YAML
from ..helper import check_file, return_filehandle, create_directories,\
    setup_logging


def prefix_fasta(target, gnm, genus, species, infra_id, key, logger):
    '''Prefix FASTA files for data store standard'''
    re_header = re.compile(r"^>(\S+)\s*(.*)")  # match header
    prefix = '{}{}'.format(genus[:3].lower(),
                           species[:2].lower())  # gensp
    new_file_raw = '{}.{}.{}.{}.genome_main.fna'.format(
        prefix, infra_id, gnm, key)
    species_dir = './{}_{}'.format(genus.capitalize(), species)
    new_file_dir = '{}/{}.{}.{}'.format(species_dir, infra_id, gnm,
                                        key)  # setup for ds
    create_directories(new_file_dir)  # make genus species dir for output
    fasta_file_path = os.path.abspath(
        '{}/{}'.format(new_file_dir, new_file_raw))
    new_fasta = open(fasta_file_path, 'w')
    name_prefix = '{}.{}.{}'.format(prefix, infra_id, gnm)
    fh = return_filehandle(target)
    with fh as gopen:
        for line in gopen:
            line = line.rstrip()
            if re_header.match(line):  # header line
                parsed_header = re_header.search(line).groups()
                logger.debug(line)
                logger.debug(parsed_header)
                hid = ''
                desc = ''
                new_header = ''
                if isinstance(parsed_header, str):  # check tuple
                    hid = parsed_header
                    new_header = '>{}.{}'.format(name_prefix, hid)
                    logger.debug(hid)
                else:
                    if len(parsed_header) == 2:  # header and description
                        hid = parsed_header[0]  # header
                        desc = parsed_header[1]  # description
                        new_header = '>{}.{} {}'.format(name_prefix, hid,
                                                        desc)
                        logger.debug(hid)
                        logger.debug(desc)
                        logger.debug(new_header)
                    else:
                        logger.error('header {} looks odd...'.format(line))
                        sys.exit(1)
                new_fasta.write(new_header + '\n')  # write new header
            else:  # sequence lines
                new_fasta.write(line + '\n')
    new_fasta.close()
    if not check_file(fasta_file_path):
        logger.error('Output file {} not found for normalize fasta'.format(
            fasta_file_path))
        sys.exit(1)  # new file not found return False
    return fasta_file_path


def type_rank(hierarchy, feature_type):
    if feature_type in hierarchy:
        return hierarchy[feature_type]['rank']
    else:
        return 1000  # no feature id, no children, no order


def update_hierarchy(hierarchy, feature_type, parent_types):
    '''Breaks input gff3 line into attributes.

       Determines feature type hierarchy using type and attributes fields
    '''
    if not parent_types:  # this feature must be a root
        if feature_type not in hierarchy:  # add to hierarchy
            hierarchy[feature_type] = {'children': [], 'parents': [],
                                       'rank': 0}
        hierarchy[feature_type]['rank'] = 1  # rank is 1 because no parents
    else:  # feature has parents
        if feature_type not in hierarchy:
            hierarchy[feature_type] = {'children': [],
                                       'parents': [], 'rank': 0}
        hierarchy[feature_type]['parents'] = parent_types
        for p in parent_types:  # set children and aprents for all p
            if hierarchy.get(p):
                if feature_type not in hierarchy[p]['children']:
                    hierarchy[p]['children'].append(feature_type)
            else:
                hierarchy[p] = {'children': [feature_type], 'parents': [],
                                'rank': 0}


def prefix_gff3(
        target,
        gnm,
        ann,
        genus,
        species,
        infra_id,
        unique_key,
        logger,
        sort_only):
    '''Prefixes GFF3 file for datastore standard.  Sorts for Tabix Indexing'''
    prefix = '{}{}'.format(genus[:3].lower(),
                           species[:2].lower())  # gensp
    key = unique_key
    target = os.path.abspath(target)
    new_file_raw = '{}.{}.{}.{}.{}.gene_models_main.gff3'.format(
        prefix, infra_id, gnm, ann, key)
    species_dir = './{}_{}'.format(genus.capitalize(), species)
    new_file_dir = '{}/{}.{}.{}.{}'.format(species_dir, infra_id, gnm,
                                           ann, key)  # setup for ds
    create_directories(new_file_dir)  # make genus species dir for output
    gff_file_path = os.path.abspath(
        '{}/{}'.format(new_file_dir, new_file_raw))
    new_gff = open(gff_file_path, 'w')
    get_id = re.compile('ID=([^;]+)')  # gff3 get id string
    get_name = re.compile('Name=([^;]+)')  # get name string
    get_parents = re.compile('Parent=([^;]+)')  # get parents
    gff3_lines = []  # will be sorted after loading and hierarchy
    sub_tree = {}  # feature hierarchy sub tree
    type_hierarchy = {}  # type hierarchy for features, will be ranked
    prefix_name = False
    fh = return_filehandle(target)  # get filehandle for target
    with fh as gopen:
        for line in gopen:
            line = line.rstrip()
            if not line:
                continue
            if line.startswith('#'):  # header
                if sort_only:  # do not prefix
                    new_gff.write('{}\n'.format(line))
                    continue
                if line.startswith('##sequence-region'):  # replace header
                    fields = re.split(r'\s+', line)
                    ref_name = '{}.{}.{}.{}'.format(prefix, infra_id,
                                                    gnm, fields[1])
                    line = re.sub(fields[1], ref_name, line)
                new_gff.write('{}\n'.format(line))
                continue
            fields = line.split('\t')
            gff3_lines.append(fields)
            feature_id = get_id.search(fields[-1])
            parent_ids = get_parents.search(fields[-1])
            if parent_ids:
                parent_ids = parent_ids.group(1).split(',')
            else:
                parent_ids = []
            if not feature_id:
                continue
            feature_id = feature_id.group(1)
            sub_tree[feature_id] = {'type': fields[2],  # parent child
                                    'parent_ids': parent_ids}
    for f in sub_tree:
        feature_type = sub_tree[f]['type']
        parent_types = []
        for p in sub_tree[f]['parent_ids']:
            parent_type = sub_tree[p]['type']
            if not parent_type:  # must have type
                logger.error('could not find type for parent {}'.format(
                    parent_type))
                sys.exit(1)  # error no type
            parent_types.append(parent_type)
        update_hierarchy(type_hierarchy, feature_type, parent_types)
    ranking = 1  # switch to stop ranking in while below
    while ranking:  # this is over-engineered for tabix indexing, but tis cool
        check = 0  # switch to indicate no features unranked
        for t in sorted(
                type_hierarchy,
                key=lambda k: type_hierarchy[k]['rank'],
                reverse=True):
            if type_hierarchy[t]['rank']:  # rank !=0
                for c in type_hierarchy[t]['children']:
                    if not type_hierarchy[c]['rank']:
                        type_hierarchy[c]['rank'] = type_hierarchy[t]['rank'] + 1
            else:  # rank == 0
                check = 1
        if not check:  # escape loop when all features are ranked
            ranking = 0
    feature_prefix = '{}.{}.{}.{}'.format(prefix, infra_id, gnm, ann)
    for l in sorted(
        gff3_lines, key=lambda x: (
            x[0], int(
            x[3]), type_rank(
                type_hierarchy, x[2]))):  # rank by chromosome, start, type_rank and stop
        if sort_only:  # do not prefix or change IDs
            l = '\t'.join(l)
            new_gff.write('{}\n'.format(l))
            continue
        l[0] = '{}.{}.{}.{}'.format(prefix, infra_id, gnm, l[0])  # rename
        l = '\t'.join(l)
        feature_id = get_id.search(l)
        feature_name = get_name.search(l)
        feature_parents = get_parents.search(l)
        if feature_id:  # if id set new id
            new_id = '{}.{}'.format(feature_prefix, feature_id.group(1))
            l = get_id.sub('ID={}'.format(new_id), l)
        if feature_name and prefix_name:  # if name and flag set new name
            new_name = '{}.{}'.format(feature_prefix, feature_name.group(1))
            l = get_name.sub('Name={}'.format(new_name), l)
        if feature_parents:  # parents set new parent ids
            parent_ids = feature_parents.group(1).split(',')
            new_ids = []
            for p in parent_ids:  # for all parents
                new_id = '{}.{}'.format(feature_prefix, p)
                new_ids.append(new_id)
            new_ids = ','.join(new_ids)  # set delimiter for multiple parents
            l = get_parents.sub('Parent={}'.format(new_ids), l)
        new_gff.write('{}\n'.format(l))
    new_gff.close()
    if not check_file(gff_file_path):  # check for output error if not found
        logger.error('Output file {} not found for normalize gff'.format(
            gff_file_path))
        sys.exit(1)  # new file not found return False
    return gff_file_path  # return path to new gff file


@click.command()
@click.option('--target',
              help='''TARGETS can be files or directories or both''')
@click.option('--gnm',
              help='''Genome Version for Normalizer.''')
@click.option('--ann',
              help='''Annotation Version for Normalizer.''')
@click.option('--genus', metavar='<STRING>',
              help='''Genus of organism input file''')
@click.option('--species', metavar='<STRING>',
              help='''Species of organism input file''')
@click.option('--infra_id', metavar='<STRING>',
              help='''Line or infraspecific identifier.  ex. A17_HM341''')
@click.option('--unique_key', metavar='<STRING, len=4>',
              help='''4 Character unique idenfier (get from spreadsheet)''')
@click.option('--sort_only', is_flag=True,
              help='''Performing sorting only (only applies to gff3 files)''')
@click.option(
    '--log_file',
    metavar='<FILE>',
    default='./normalizer_prefix.log',
    help='''File to write log to. (default:./normalizer_prefix.log)''')
@click.option(
    '--log_level',
    metavar='<LOGLEVEL>',
    default='INFO',
    help='''Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)''')
def cli(
        target,
        gnm,
        ann,
        genus,
        species,
        infra_id,
        unique_key,
        sort_only,
        log_file,
        log_level):
    '''Prefixing for Datastore standard.

       full prefix = gensp.infra_id.gnm.[ann]
       file prefix = gensp.infra_id.gnm.[ann].unique_key

       If the File is a GFF3 it will be sorted.

       A directory will be created in ./Genus_species containing corretly
       named subdirectories and files.

       You can run all of the other steps on the results of this command.
    '''
    file_types = ['fna', 'fasta', 'fa', 'gff', 'gff3', 'ADD MORE']
    fasta = ['fna', 'fasta', 'fa']
    gff3 = ['gff', 'gff3']
    logger = setup_logging(log_file, log_level)
    if not (target and gnm and genus and species and infra_id and unique_key):
        logger.error(
            '--target, --species, --genus, --gnm, --infra_id, and --unique_key are required')
        sys.exit(1)
    target = os.path.abspath(target)
    gnm = 'gnm{}'.format(gnm)  # format user provided argument for gnm
    ann = 'ann{}'.format(ann)  # format user provided argument for ann
    file_attributes = target.split('.')
    new_file = False
    if len(file_attributes) < 2:  # file must have an extension
        error_message = '''Target {} does not have a type or attributes.  File must end in either gz, bgz, fasta, fna, fa, gff, or gff3.'''.format(
            target)
        logger.error(error_message)
        sys.exit(1)
    file_type = file_attributes[-1]
    if file_type == 'gz' or file_type == 'bgz':  # if compressed get upstream
        file_type = file_attributes[-2]
    if file_type not in file_types:  # must be a type in file_types
        logger.error('File {} is not a type in {}'.format(target,
                                                          file_types))
        sys.exit(1)
    if file_type in fasta:
        logger.info('Target is a FASTA file prefixing...')
        new_file = prefix_fasta(target, gnm, genus, species, infra_id,
                                unique_key, logger)  # returns new file path
        logger.info('Prefixing done, final file: {}'.format(new_file))
    if file_type in gff3:
        if sort_only:
            logger.info('Target is a gff3 file sorting ONLY...')
        else:
            logger.info('Target is a gff3 file prefixing and sorting...')
        new_file = prefix_gff3(
            target,
            gnm,
            ann,
            genus,
            species,
            infra_id,
            unique_key,
            logger,
            sort_only)  # returns new file path
        logger.info('Prefixing done, final file: {}'.format(new_file))
    if not new_file:  # no new file was generated, must have errored
        logger.error('Normalizer FAILED.  See Log.')
        sys.exit(1)
    return new_file



#!/usr/bin/env python

import os
import sys
import logging
import re
import subprocess
import gzip
import click
from ruamel.yaml import YAML, RoundTripDumper
from ..helper import check_file, return_filehandle, create_directories,\
    check_subprocess_dependencies, setup_logging


def read_yaml(template):
    '''Reads yaml file and checks for valid yaml'''
    yaml = YAML()
    target_yaml = yaml.load(open(template, 'rt'))
    return yaml, target_yaml


def generate_readme(template, attributes, logger):
    '''Writes target YAML file'''
    yaml, my_yaml = read_yaml(template)
    my_readme = '{}/README.{}.yaml'.format(attributes['target_dir'],
                                           attributes['key'])
    identifier = attributes['key']
    sci_name = '{} {}'.format(attributes['genus'].capitalize(),
                              attributes['species'].lower())
    sci_name_abr = attributes['gensp'].lower()
    genotype = attributes['infra_id']
    my_yaml['identifier'] = identifier  # set key
    my_yaml['genotype'] = genotype  # set infraspecific id
    my_yaml['scientific_name'] = sci_name  # Genus species
    my_yaml['scientific_name_abbrev'] = sci_name_abr  # gensp
    readme_handle = open(my_readme, 'w')
    yaml.dump(my_yaml, readme_handle)
#    for m in my_yaml:
    readme_handle.close()


@click.command()
@click.option('--target', type=str,
              help='''Formatted Directory from Previous Normalizer Steaps''')
@click.option('--template', type=str,
              help='''YAML file to be used as the template for the README''')
@click.option(
    '--log_file',
    metavar='<FILE>',
    default='./normalizer_index.log',
    help='''File to write log to. (default:./normalizer_index.log)''')
@click.option(
    '--log_level',
    metavar='<LOGLEVEL>',
    default='INFO',
    help='''Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)''')
def cli(target, template, log_file, log_level):
    '''Determines what typae of index to apply to input target'''
    logger = setup_logging(log_file, log_level)
    if not (target and template):
        logger.error('--target and --template arguments are required')
        sys.exit(1)
    target = os.path.abspath(target)  # get full path
    if not os.path.isdir(target):
        logger.error('target dmust be a directory')
        sys.exit(1)
    organism_dir = os.path.basename(os.path.dirname(target))
    target_dir = os.path.basename(target)
    organism_attributes = organism_dir.split('_')  # Genus_species
    target_attributes = target_dir.split('.')
    if len(organism_attributes) != 2:
        logger.error(
            'Parent directory {} is not Genus_species'.format(organism_dir))
        sys.exit(1)
    if len(target_attributes) < 3:
        logger.error(
            'Target directory {} is not delimited correctly'.format(target_dir))
        sys.exit(1)
    # abbreviation in yaml README file
    gensp = '{}{}'.format(
        organism_attributes[0][:3], organism_attributes[1][:2])
    attributes = {'genus': organism_attributes[0],
                  'species': organism_attributes[1],
                  'gensp': gensp,
                  'key': target_attributes[-1],
                  'infra_id': target_attributes[0],
                  'target_dir': target}
#    print('{}\t{}\t{}'.format(template, attributes, logger))
    generate_readme(template, attributes, logger)  # write readme template



#!/usr/bin/env python

import os
import sys
import logging
import re
import subprocess
import hashlib
import requests
import json
import importlib
import importlib.util
from glob import glob
from sequencetools.tools.basic_fasta_stats import basic_fasta_stats
from . import specification_checks
from .helper import check_file, return_filehandle
from .metrics import count_gff_features


class Detector:
    '''Class to detect datastore file incongruencies with

       https://github.com/LegumeFederation/datastore
    '''

    def __init__(self, target, **kwargs):
        '''Check for logger, check for gt'''
        subprocess.check_call('which gt', shell=True)  # check for gt in env
        self.options = kwargs
        self.checks = {}  # object that determines which checks are skipped
        self.checks['genome_main'] = kwargs.get('disable_genome_main', True)
        self.checks['gene_models_main'] = kwargs.get(
            'disable_gene_models_main', True)
        self.checks['perform_gt'] = kwargs.get('disable_gt', True)
        self.checks['fasta_headers'] = kwargs.get(
            'disable_fasta_headers', True)
        self.no_nodes = kwargs.get('no_nodes', False)
        self.canonical_types = ['genome_main',
                                'protein_primaryTranscript',
                                'protein',
                                'gene_models_main',
                                'ADDMORESTUFF']  # types for detector
        self.canonical_parents = {
            'genome_main': None,
            'gene_models_main': 'genome_main',
            'protein_primaryTranscript': 'gene_models_main',
            'protein': 'gene_models_main'}
        self.rank = {'genome_main': 0, 'gene_models_main': 1, 'protein': 2,
                     'protein_primaryTranscript': 2}
        self.log_level = kwargs.get('log_level', 'INFO')
        self.log_file = kwargs.get('log_file', './incongruencies.log')
        self.output_prefix = kwargs.get('output_prefix', './incongruencies')
        self.write_me = {}
        self.passed = {}  # dictionary of passing names
        self.target_objects = {}  # store all target pairings self.get_targets
        self.fasta_ids = {}
        self.reporting = {}
        self.node_data = {}  # nodes for DSCensor
        self.target = os.path.abspath(target)
        self.target_readme = ''
        self.domain = 'https://legumeinfo.org/data/public'
        self.setup_logging()  # setup logging sets self.logger
        logger = self.logger
        self.get_target_type()  # sets self.target_name and self.target_type
        if not self.target_type:  # target type returned False not recognized
            logger.error('Target type not recognized for {}'.format(
                self.target))
            sys.exit(1)
        # if not os.environ.get('BUSCO_ENV_FILE', None) and not kwargs.get('no_busco'):
        #    logger.error('''
        #        BUSCO_ENV_File Must Be set, please export
#
#                This is used to source an Environment for BUSCO as the BUSCO
#                Config Does Still Requires AUGUSTUS Environment variables be set
#                ''')
        logger.info('Target type looks like {}'.format(self.target_type))
        self.get_targets()
#        logger.debug(''.format(self.target_objects))
        logger.info('Performing Checks for the Following:\n')
        for t in self.target_objects:  # for each object set validate
            logger.debug('{}'.format(t))
            logger.debug('{}'.format(self.target_objects[t]))
            logger.info('Parent {}:'.format(t))
            count = 0
            set_primary = ''
            primary = False
            for c in self.target_objects[t]['children']:
                logger.info('Child {}'.format(c))
                if self.target_objects[t]['children'][c]['type'] == 'protein_primaryTranscript':
                    primary = True
                if self.target_objects[t]['children'][c]['type'] == 'protein':
                    count += 1
                    set_primary = c
            if count == 1 and not primary:
                self.target_objects[t]['children'][set_primary]['type'] = 'protein_primaryTranscript'
                self.target_objects[t]['children'][set_primary]['node_data']['canonical_type'] = 'protein_primaryTranscript'
            if count > 1 and not primary:
                logger.error(
                    'Multiple protein files found for {} one must be renamed to primary'.format(t))
                sys.exit(1)
#            for c in self.target_objects[t]['children']:
#                logger.info('{}'.format(c))
        logger.info('Initialized Detector\n')

    def setup_logging(self):
        '''Return logger based on log_file and log_level'''
        log_level = getattr(logging, self.log_level.upper(), logging.INFO)
        msg_format = '%(asctime)s|%(name)s|[%(levelname)s]: %(message)s'
        logging.basicConfig(format=msg_format, datefmt='%m-%d %H:%M',
                            level=log_level)
        log_handler = logging.FileHandler(self.log_file, mode='w')
        formatter = logging.Formatter(msg_format)
        log_handler.setFormatter(formatter)
        logger = logging.getLogger('detect_incongruencies')
        logger.addHandler(log_handler)
        self.logger = logger

    def get_target_type(self):
        '''Determines what type of target this is, is it a file or directory?

           if its a directory, is it an organism directory or a data directory?
        '''
        logger = self.logger
        target_name = os.path.basename(self.target)
        self.target_name = target_name
        self.target_type = False
        logger.debug(self.target_name)
        if target_name.endswith('.gz'):  # all datastore files end in .gz
            self.target_type = 'file'
        elif(len(target_name.split('_')) == 2 and
             len(target_name.split('.')) < 3):
            self.target_type = 'organism_dir'  # will always be Genus_species
        elif len(target_name.split('.')) >= 3:  # standard naming minimum
            self.target_type = 'data_dir'

    def get_targets(self):
        '''Gets and discovers target files relation to other files

           if the target is a directory, the program will discover

           all related files that can be checked
        '''
        logger = self.logger
        target = self.target
        target_type = self.target_type
        if target_type == 'file':  # starting with a file
            self.add_target_object()
            return
        elif target_type == 'data_dir' or target_type == 'organism_dir':
            self.get_all_files()  # works for both data and organism
            return
        # elif target_type == 'organism_dir':  # entire organism
        #    self.get_all_files()
        #    return

    def get_all_files(self):
        '''Walk down filetree and recursively return all files

           These will be checked using add_target_object
        '''
        logger = self.logger
        target = self.target
        for root, directories, filenames in os.walk(target):
            for filename in filenames:  # we only care about the files
                my_target = os.path.join(root, filename)  # make path
                logger.debug('Checking file {}'.format(my_target))
                self.target = my_target  # set target
                self.target_name = os.path.basename(self.target)
                self.add_target_object()  # add target if canonical

    def get_target_file_type(self, target_file):
        '''Determines if file is fasta, gff3, vcf, etc'''
        file_type = target_file.split('.')[-2].lower()
        if file_type == 'fna' or file_type == 'faa' or file_type == 'fasta':
            file_type = 'fasta'
        elif file_type == 'gff' or file_type == 'gff3':
            file_type = 'gff3'
        else:  # add more
            return False
        return file_type

    def add_target_object(self):
        '''Uses parent child logic to create a datastructure for objects'''
        logger = self.logger
        target = self.target
        logger.debug(target)
        if not check_file(target):
            logger.error('Could not find file: {}'.format(target))
            sys.exit(1)
        target_attributes = self.target_name.split('.')
        if len(target_attributes) < 3 or self.target_name[0] == '_':
            logger.debug('File {} does not seem to have attributes'.format(
                target))
            return
        canonical_type = target_attributes[-3]  # check content type
        if canonical_type not in self.canonical_types:  # regject
            logger.debug('Type {} not recognized in {}.  Skipping'.format(
                canonical_type,
                self.canonical_types))
            return
        organism_dir_path = os.path.dirname(os.path.dirname(target))  # org dir
        organism_dir = os.path.basename(
            os.path.dirname(
                os.path.dirname(target)))  # org dir
        target_dir = os.path.basename(os.path.dirname(target))
        genus = organism_dir.split('_')[0].lower()
        species = organism_dir.split('_')[1].lower()
        target_format = target_attributes[-2]  # get gff, fna, faa all gx
        target_key = target_attributes[-4]  # get key
        target_ref_type = self.canonical_parents[canonical_type]
        logger.debug('Getting target files reference if necessary...')
        file_type = self.get_target_file_type(self.target_name)
        file_url = '{}/{}/{}/{}'.format(self.domain, organism_dir,
                                        target_dir, self.target_name)
        target_node_object = {
            'filename': self.target_name,
            'filetype': file_type,
            'canonical_type': canonical_type,
            'url': file_url,
            'counts': '',
            'genus': genus,
            'species': species,
            'origin': 'LIS',
            'infraspecies': target_attributes[1],
            'derived_from': [],
            'child_of': []}
        if len(target_attributes) > 7 and target_ref_type:  # check parent
            logger.debug('Target Derived from Some Reference Searching...')
            ref_glob = '{}/{}*/*{}.*.gz'.format(organism_dir_path,
                                                '.'.join(target_attributes[1:3]),
                                                target_ref_type)
            if self.rank[canonical_type] > 1:  # feature has a subtype
                ref_glob = '{}/{}*/*{}.*.gz'.format(organism_dir_path,
                                                    '.'.join(target_attributes[1:4]),
                                                    target_ref_type)
            my_reference = self.get_reference(ref_glob)
            if my_reference not in self.target_objects:  # new parent
                parent_name = os.path.basename(my_reference)
                file_type = self.get_target_file_type(parent_name)
                organism_dir = os.path.basename(
                    os.path.dirname(os.path.dirname(my_reference)))
                ref_dir = os.path.basename(os.path.dirname(my_reference))
                file_url = '{}/{}/{}/{}'.format(self.domain, organism_dir,
                                                ref_dir, parent_name)
                ref_node_object = {
                    'filename': parent_name,
                    'filetype': file_type,
                    'canonical_type': target_ref_type,
                    'url': file_url,
                    'counts': '',
                    'genus': genus,
                    'species': species,
                    'origin': 'LIS',
                    'infraspecies': target_attributes[1],
                    'derived_from': [],
                    'child_of': []}
                target_node_object['child_of'].append(parent_name)
                target_node_object['derived_from'].append(parent_name)
                self.target_objects[my_reference] = {
                    'type': target_ref_type,
                    'node_data': ref_node_object,
                    'readme': '',
                    'children': {}}
                self.target_objects[my_reference]['children'][target] = {
                    'node_data': target_node_object,
                    'type': canonical_type}
            else:  # the parent is already in the data structure add child
                if target not in self.target_objects[my_reference]['children']:
                    parent_name = os.path.basename(my_reference)
                    target_node_object['child_of'].append(parent_name)
                    target_node_object['derived_from'].append(parent_name)
                    self.target_objects[my_reference]['children'][target] = {
                        'node_data': target_node_object,
                        'type': canonical_type}
        else:  # target is a reference
            if target_ref_type:
                logger.error('Reference was not found or file has <=7 fields')
                sys.exit(1)
            logger.debug('Target has no Parent, it is a Reference')
            if target not in self.target_objects:
                self.target_objects[target] = {'type': canonical_type,
                                               'node_data': target_node_object,
                                               'children': {}}

    def get_reference(self, glob_target):
        '''Finds the FASTA reference for some prefix'''
        logger = self.logger
        if len(glob(glob_target)) > 1:  # too many references....?
            logger.error('Multiple references found {}'.format(glob_target))
            sys.exit(1)
        reference = glob(glob_target)
        if not reference:  # if the objects parent could not be found
            logger.error('Could not find ref glob: {}'.format(glob_target))
            sys.exit(1)
        reference = glob(glob_target)[0]
        if not os.path.isfile(reference):  # if cannot find reference file
            logger.error('Could not find main target {}'.format(reference))
            sys.exit(1)
        logger.debug('Found reference {}'.format(reference))
        return reference

    def write_node_object(self):
        '''Write object node for loading into DSCensor

           file name is object name from json
        '''
        my_name = self.write_me['filename']
        if self.write_me['canonical_type'] == 'genome_main':
            self.write_me['counts'] = basic_fasta_stats(self.target, 10, False)
        elif self.write_me['canonical_type'] == 'gene_models_main':
            self.write_me['counts'] = count_gff_features(self.target)
        my_file = open('./{}.json'.format(my_name), 'w')
        my_file.write(json.dumps(self.write_me))
        my_file.close()

    def run_busco(self, mode, file_name):
        '''Runs BUSCO using BUSCO_ENV_FILE env var and input

           and outputs to file_name
        '''
        logger = self.logger  # set logging
        node_data = self.node_data  # get current targets nodes
        busco_parse = re.compile(
            r'C:(.+)\%\[S:(.+)\%,D:(.+)\%\],F:(.+)\%,M:(.+)\%,n:(\d+)')
        output = '{}.busco'.format('.'.join(file_name.split('.')[:-2]))
        cmd = 'run_BUSCO.py --mode {} --lineage {}'.format(mode, 'lineage')
        outdir = './run_{}'.format(output)  # output from BUSCO
        short_summary = glob(outdir + '/short_summary*.busco.txt')
        if not short_summary:
            logger.error('BUSCO short summary not found for {}'.format(outdir))
            sys.exit(1)
        short_summary = short_summary[0]
        with open(short_summary) as fopen:
            for line in fopen:
                line = line.rstrip()
                if line.startswith('#') or not line:
                    continue
                if line.startswith('\tC:'):
                    line = line.replace('\t', '')
                    percentages = busco_parse.search(line)  # read summary
                    total = int(percentages.group(6))
                    complete = float(percentages.group(1))
                    frag = float(percentages.group(4))
                    single = float(percentages.group(2))
                    duplicate = float(percentages.group(3))
                    missing = float(percentages.group(5))
                    node_data['busco'] = {'total_buscos': total,  # node BUSCO
                                          'complete_buscos': complete,
                                          'fragmented_buscos': frag,
                                          'single_copy_buscos': single,
                                          'duplicate_buscos': duplicate,
                                          'missing_buscos': missing}

    def detect_incongruencies(self):
        '''Detects all incongruencies in self.target_objects

           report all incongruencies to self.output_prefix

           Probably refactor this at some point to use a self.dict

           This is to allow just iteration and lookup to occur and remove

           basically all this conditional foo
        '''
        logger = self.logger
        targets = self.target_objects  # get objects from class init
        no_nodes = self.no_nodes  # if true, no nodes for DSCensor written
        for reference in sorted(targets,
                                key=lambda k: self.rank[targets[k]['type']]):
            #            logger.info('HERE {}'.format(reference))
            if reference not in self.passed:
                self.passed[reference] = 0
                self.target = reference
                ref_method = getattr(specification_checks,  # reads checks from spec
                                     targets[reference]['type'])  # type ex genome_main
                if not ref_method:  # if the target isnt in the hierarchy continue
                    logger.debug('Check for {} does not exist'.format(
                        targets[reference]['type']))
                    continue
                logger.debug(ref_method)
                my_detector = ref_method(self, **self.options)
                passed = my_detector.run()
                if passed:  # validation passed writing object node for DSCensor
                    self.passed[reference] = 1
                    self.node_data = targets[reference]['node_data']
    #                if targets[reference]['type'] == 'genome_main':  # BUSCO
    #                    file_name = targets[reference]['node_data']['filename']
    #                    if not self.options.get('no_busco'):
    #                        self.run_busco('genome', file_name)
                    if not no_nodes:
                        logger.info(
                            'Writing node object for {}'.format(reference))
                        # dscensor node
                        self.write_me = targets[reference]['node_data']
                        self.write_node_object()  # write node for dscensor loading
                logger.debug('{}'.format(targets[reference]))
            if self.target_objects[reference]['children']:  # process children
                children = self.target_objects[reference]['children']
                for c in children:
                    if c in self.passed:
                        logger.debug('Child {} Already Passed'.format(c))
                        continue

                    self.passed[c] = 0
#                    logger.info('HERE child {}'.format(c))
                    logger.info('Performing Checks for {}'.format(c))
                    self.target = c
                    child_method = getattr(specification_checks,  # check for spec
                                           children[c]['type'])  # exgene_models_main
                    if not child_method:
                        logger.warning('Check for {} does not exist'.format(
                            children[c]['type']))
                        continue
                    logger.debug(child_method)
                    my_detector = child_method(self, **self.options)
                    passed = my_detector.run()
                    if passed:  # validation passed writing object node for DSCensor
                        self.passed[c] = 1
                        if not no_nodes:
                            logger.info('Writing node object for {}'.format(c))
                            self.write_me = children[c]['node_data']
                            self.write_node_object()
                    logger.debug('{}'.format(c))



#!/usr/bin/env python

import os
import sys
import gzip
import click
import logging
from . import Detector


@click.command()
@click.option('--target', multiple=True,
              help='''TARGETS can be files or directories or both''')
@click.option('--no_busco', is_flag=True,
              help='''Disable BUSCO''')
@click.option('--no_nodes', is_flag=True,
              help='''Disable DSCensor Stats and Node Generation''')
@click.option('--log_file', metavar='<FILE>',
              default='./bionorm.log',
              help='''File to write log to. (default:./bionorm.log)''')
@click.option(
    '--log_level',
    metavar='<LOGLEVEL>',
    default='INFO',
    help='''Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)''')
# @click.option('--normalize', is_flag=True,
#             help='''Normalizes provided files.
# Incongruencies in FASTA will be corrected if the provided genome name
# passes checks.
# The gff file will be tidied if it fails gff3 validation in gt:
#    gt gff3 -sort -tidy -retainids input.gff3 > out.gff3
# ''')
def cli(target, no_busco, no_nodes, log_file, log_level):
    '''incongruency_detector:

        Detect Incongruencies with LIS Data Store Standard
    '''
    if not target:
        print('Must specify at least one --target.  Run with --help for usage')
        sys.exit(1)
    data_store_home = os.path.dirname(os.path.dirname(
        os.path.abspath(__file__)))
    templates_dir = '{}/templates'.format(data_store_home)
    readme_template = '{}/template__README.KEY.yml'.format(templates_dir)
    options = {'log_level': log_level, 'log_file': log_file,
               'no_busco': no_busco, 'no_nodes': no_nodes,
               'readme_template': readme_template}
    for t in target:
        detector = Detector(t, **options)  # initializers
        detector.detect_incongruencies()  # class method runs all detection methods


if __name__ == '__main__':
    cli()



#!/usr/bin/env python

import os
import sys
import logging
import re
import subprocess
import gzip
import click
from ruamel.yaml import YAML
from ..helper import check_file, return_filehandle, create_directories,\
    check_subprocess_dependencies, setup_logging


def index_fasta(fasta):
    '''Indexes the supplied fasta file using bgzip and samtools'''
    compressed_fasta = '{}.gz'.format(fasta)  # will now have gz after bgzip
    if not check_file(compressed_fasta):
        cmd = 'bgzip -f --index {}'.format(fasta)  # bgzip command
        subprocess.check_call(cmd, shell=True)
    cmd = 'samtools faidx {}'.format(compressed_fasta)  # samtools faidx
    subprocess.check_call(cmd, shell=True)
    return compressed_fasta  # return new compressed and indexed gff


def index_gff3(gff3):
    '''Indexes the supplied gff3 file using bgzip and tabix'''
    compressed_gff = '{}.gz'.format(gff3)  # will now have gz after bgzip
    if not check_file(compressed_gff):
        cmd = 'bgzip -f {}'.format(gff3)  # bgzip command
        subprocess.check_call(cmd, shell=True)
    cmd = 'tabix -p gff {}'.format(compressed_gff)  # tabix index command
    standard_outval = subprocess.call(cmd.split(' '))
    if standard_outval:
        # tabix index command
        cmd = 'tabix --csi -p gff {}'.format(compressed_gff)
        subprocess.check_call(cmd, shell=True)
    return compressed_gff  # return new compressed and indexed gff


@click.command()
@click.option('--target', type=str,
              help='''TARGETS can be files or directories or both''')
@click.option(
    '--log_file',
    metavar='<FILE>',
    default='./normalizer_index.log',
    help='''File to write log to. (default:./normalizer_index.log)''')
@click.option(
    '--log_level',
    metavar='<LOGLEVEL>',
    default='INFO',
    help='''Log level: DEBUG, INFO, WARNING, ERROR, CRITICAL (default:INFO)''')
def cli(target, log_file, log_level):
    '''Determines what type of index to apply to input target'''
    check_subprocess_dependencies()
    logger = setup_logging(log_file, log_level)
    if not target:
        logger.error('--target argument required')
        sys.exit(1)
    target = os.path.abspath(target)
    file_types = ['fna', 'fasta', 'fa', 'gff', 'gff3', 'faa', 'ADD MORE']
    fasta = ['fna', 'fasta', 'fa', 'faa']
    gff3 = ['gff', 'gff3']
    file_attributes = target.split('.')
    new_file = False
    if len(file_attributes) < 2:
        error_message = '''Target {} does not have a type or attributes.  File must end in either gz, bgz, fasta, fna, fa, gff, or gff3.'''.format(
            target)
        logger.error(error_message)
        sys.exit(1)
    file_type = file_attributes[-1]
    if file_type == 'gz' or file_type == 'bgz':
        file_type = file_attributes[-2]
        logger.error(
            'Uncompress file for indexing.  Compression is done specifically for each index type')
        sys.exit(1)
    if file_type not in file_types:
        logger.error('File {} is not a type in {}'.format(target,
                                                          file_types))
        sys.exit(1)
    if file_type in fasta:
        logger.info('Target is a FASTA file indexing...')
        new_file = index_fasta(target)
        logger.info('Indexing done, final file: {}'.format(new_file))
    if file_type in gff3:
        logger.info('Target is a gff3 file indexing...')
        new_file = index_gff3(target)
        logger.info('Indexing done, final file: {}'.format(new_file))
    if not new_file:
        logger.error('Indexing FAILED.  See Log.')
        sys.exit(1)
    return new_file



# -*- coding: utf-8 -*-
'''Global constants and common helper functions.
'''
# standard library imports
import locale
import logging
import sys
from datetime import datetime
from pathlib import Path

# 3rd-party modules
import click

#
# global constants
#
PROGRAM_NAME = 'bionorm'
VERSION = "0.1.blah"

DEFAULT_FIRST_N = 0  # only process this many records

CONFIG_FILE_ENVVAR = 'BIONORM_CONFIG_FILE_PATH'
#
# global logger object
#
logger = logging.getLogger(PROGRAM_NAME)
#
# set locale so grouping works
#
for localename in ['en_US', 'en_US.utf8', 'English_United_States']:
    try:
        locale.setlocale(locale.LC_ALL, localename)
        break
    except BaseException:
        continue
#
# Classes and instances of classes
#
class PersistentConfigurationObject(object):
    '''Defines a persistent configuration object

    Attributes:
        :config_dict: Dictionary of configuration parameters.
        :name: Configuration filename.
        :path: Configuration filepath (absolute).
    '''

    def __init__(self, config_dir=None, name='config.yaml'):
        '''Inits the configuration dictionary

        Reads a configuration file from a subdirectory of the
        current working directory.  If that file isn't found,
        then searches in the system-specific configuration directory.
        If that file isn't found either, creates a new file in the
        directory specified by the location parameter.
        '''
        self.name = name
        self._default_dict = {'version': VERSION,
                              'simplicity_object_label': None,
                              'plot_type': 'pdf',
                              'sets': [],
                              'summary': {'dir': None,
                                          'label': None},
                              }

        self._default_path = Path(
            (click.get_app_dir(PROGRAM_NAME) + '/' + self.name))
        self._cwd_path = Path('.' + '/.' + PROGRAM_NAME + '/' + self.name)
        if config_dir is not None:
            self.path = self._get_path_from_dir(config_dir)
        elif self._cwd_path.is_file():
            self.path = self._cwd_path
        else:
            self.path = self._default_path
        if not self.path.exists():
            self.config_dict = {}
            self.path = None
        else:
            with self.path.open('rt') as f:
                self.config_dict = yaml.safe_load(f)

    def _get_path_from_dir(self, dir):
        return Path(
            str(dir) +
            '/.' +
            PROGRAM_NAME +
            '/' +
            self.name).expanduser()

    def _update_config_dict(self):
        '''Update configuration dictionary if necessary
        :param config_dict: Configuration dictionary.
        :return: Updated configuration dictionary.
        '''
        try:
            if self.config_dict['version'] != VERSION:
                # Do whatever updates necessary, depending on version.
                # For now, nothing needs to be done.
                self.config_dict['version'] = VERSION
        except KeyError:
            logger.warning('Initializing config file "%s"',
                           self.path)
            self.config_dict = self._default_dict

    def write_config_dict(self, config_dict=None, dir=None):
        '''Writes a YAML configuration dictionary
        :param config_dict: Configuration dictionary
        :return: None
        '''
        if dir is None or dir is '':
            if self.path is None:
                self.path = self._default_path
        elif dir is '.':
            self.path = self._cwd_path
        else:
            self.path = self._get_path_from_dir(dir)

        if config_dict == {}:
            self.config_dict = self._default_dict
        elif config_dict is not None and config_dict != self.config_dict:
            self.config_dict = config_dict
            self._update_config_dict()

        if not self.path.parent.exists():
            # create parent directory
            logger.debug('Creating config file directory "%s"',
                         self.path.parent)
            try:
                self.path.parent.mkdir(parents=True)
            except OSError:
                logger.error('Unable to create parent directory "%s".',
                             self.path.parent)
                sys.exit(1)
        if not self.path.parent.is_dir():
            logger.error('Path "%s" exists, but is not a directory.',
                         self.path.parent)
            sys.exit(1)
        if not self.path.exists():
            logger.debug('Creating config file "%s"', self.path)
            try:
                self.path.touch()
            except OSError:
                logger.error('Path "%s" is not writable.', self.path)
                sys.exit(1)
        with self.path.open(mode='wt') as f:
            yaml.dump(self.config_dict, f)


config_obj = PersistentConfigurationObject()


#
# helper functions used in multiple places
#
def get_user_context_obj():
    '''Returns the user context, containing logging and configuration data.

    :return: User context object (dict)
    '''
    return click.get_current_context().obj


def to_str(seq):
    '''Decode bytestring if necessary.

    :param seq: Input bytestring, string, or other sequence.
    :return: String.
    '''
    if isinstance(seq, bytes):
        value = seq.decode('utf-8')
    elif isinstance(seq, str):
        value = seq
    else:
        value = str(seq)
    return value


def to_bytes(seq):
    '''Encode or convert string if necessary.

    :param seq: Input string, bytestring, or other sequence.
    :return: Bytestring.
    '''
    if isinstance(seq, str):
        value = seq.encode('utf-8')
    elif isinstance(seq, bytes):
        value = seq
    else:
        value = bytes(seq)
    return value



# -*- coding: utf-8 -*-
'''Implements commands related to configuration parameters
'''
# standard library imports
import sys
# third-party imports
import click
# module imports
from . import cli
from .common import *

# private context function
_ctx = click.get_current_context

@cli.command()
def show_config():
    '''Print location and contents of config file.

        Example:
            bionorm -v show_config
    '''
    global config_obj
    if config_obj.config_dict == {}:
        logger.info('No configuration file was found.')
    else:
        logger.info('Configuration file path is "%s".', config_obj.path)
        for key in config_obj.config_dict.keys():
            logger.info('  %s: %s', key, config_obj.config_dict[key])


@cli.command()
@click.argument('plot_type', type=str, nargs=-1)
def set_plot_type(plot_type):
    '''Define label associated with a set.
    '''
    global config_obj
    plot_types = ['svg', 'jpg', 'ps', 'pdf', 'svgr', 'png', 'tiff', 'eps']
    if len(plot_type) == 0:
        logger.info('Supported plot types are:')
        for typename in plot_types:
            logger.info('   %s', typename)
    elif len(plot_type) > 1:
        logger.error('Only one argument for set_plot_type is allowed.')
        sys.exit(1)
    elif plot_type[0] not in plot_types:
        logger.error('Plot type "%s" is not defined.', plot_type[0])
        logger.info('Supported plot types are:')
        for typename in plot_types:
            logger.info('   %s', typename)
        sys.exit(1)
    else:
        config_obj.config_dict['plot_type'] = plot_type[0]
        logger.info('Plot type is now %s.', plot_type[0])
        config_obj.write_config_dict()


@cli.command()
@click.argument('dir', type=str, default='')
def init_config_file(dir):
    '''Initialize a configuration file.

    :param dir: Optional directory in which to initialize the file.
    If not present, the system-dependent default application directory
    will be used.  If this argument is '.', then the current working
    directory will be used.  This argument accepts tilde expansions.
    '''
    global config_obj
    config_obj.write_config_dict(dir=dir, config_dict={})



# -*- coding: utf-8 -*-
'''bionorm -- normalize and validate genomic data files.

This tool is used to ingest data into the Legume Federation Data Store
(https://www.legumefederation.org/en/data-store/) and is useful anytime
genomes must be formatted and checked for consistency.
'''

# standard library imports
import warnings
from pkg_resources import iter_entry_points
# third-party imports
import click
import coverage
from pkg_resources import iter_entry_points
from click_plugins import with_plugins

# module defs
from .common import logger, PROGRAM_NAME
from .cli import Logging_CLI_Builder

# global constants
DEFAULT_FIRST_N = 0

# start coverage
coverage.process_startup()

# define parser options
parser_options = [{'args': ('--progress',),
                    'kwargs': {'is_flag': True,
                               'show_default': True,
                               'default': False,
                               'help': 'Show a progress bar.'}},
                  {'args': ('--first_n',),
                   'kwargs': {'default': 0,
                              help: 'Process only this many records. [default: all]'}},
                  {'args': ('--warnings_as_errors', '-e'),
                            'kwargs': {'is_flag': True,
                                       'show_default': True,
                                       'default': False,
                                       'help': 'Warnings cause exceptions.'}}
                  ]
# define custom CLI check function, parameters must be keyworded
def bionorm_check(warnings_as_errors=False, **others):
    if warnings_as_errors:
        logger.warn('Runtime warnings (e.g., from pandas) will cause exceptions')
        warnings.filterwarnings('error')

# define the CLI
cli_builder = Logging_CLI_Builder(PROGRAM_NAME,
                                  logger,
                                  global_options_list=parser_options)

# create CLI
@with_plugins(iter_entry_points(PROGRAM_NAME + '.cli_plugins'))
@click.group(epilog=cli_builder.author + ' <'
                    + cli_builder.email + '>.  '
                    + cli_builder.copyright)
@click.option('-v', '--verbose', is_flag=True, show_default=True,
              default=False, help='Log debugging info to stderr.')
@click.option('-q', '--quiet', is_flag=True, show_default=True,
              default=False, help='Suppress logging to stderr.')
@click.option('--log/--no_log', is_flag=True, show_default=True,
              default=True, help='Log to file.')
@click.option('--progress', is_flag=True, show_default=True,
              default=False, help='Show a progress bar.')
@click.option('--first_n', default=DEFAULT_FIRST_N,
              help='Process only this many records. [default: all]')
@click.option('--warnings_as_errors', '-e', is_flag=True, show_default=True,
              default=False, help='Warnings cause exceptions.')
@click.version_option(version=cli_builder.version, prog_name=PROGRAM_NAME)
@cli_builder.init_dual_logger()
@cli_builder.init_user_context_obj(extra_args=['progress', 'first_n'])
def cli(verbose,
          quiet,
          log,
          **kwargs):
    """bionorm -- normalize and verify genomic data files

        If COMMAND is present, and --no_log was not invoked,
        a log file named bionorm-COMMAND.log
        will be written in the ./logs/ directory.
    """
    bionorm_check(**kwargs)

cli_builder.set_cli_func(cli)

# Define some cli-related commands
test_log = cli_builder.test_log_func()
show_context = cli_builder.show_context_func()

# import other functions
from .config import show_config, init_config_file




# -*- coding: utf-8 -*-
'''Define a cli with dual logging to stderr and file
'''

# standard library imports
import functools
import logging
import sys
from datetime import datetime
from pathlib import Path
# third-party imports
import click
try:
    from importlib.metadata import distribution # python 3.8 and above
except ModuleNotFoundError:
    from importlib_metadata import distribution

# module defs
from .common import DEFAULT_FIRST_N

# global constants
DEFAULT_FILE_LOGLEVEL = logging.DEBUG
DEFAULT_STDERR_LOGLEVEL = logging.INFO
STARTTIME = datetime.now()

class CleanInfoFormatter(logging.Formatter):
    '''very clean logging formatter for terminal output
    '''
    def __init__(self, fmt='%(levelname)s: %(message)s'):
        logging.Formatter.__init__(self, fmt)

    def format(self, record):
        if record.levelno == logging.INFO:
            return record.getMessage()
        return logging.Formatter.format(self, record)


class Click_Multi(object):

    def __init__(self, option_list):
        self.option_list = option_list

def composed(self, *decs):
        def deco(f):
            for dec in reversed(decs):
                f = dec(f)
            return f
        return deco


def click_multi(func):
        return composed(
            click.option('--progress', is_flag=True, show_default=True,
                         default=False, help='Show a progress bar.')
            #click.option(*self.global_options_list[0]['args'],
            #             **self.global_options_list[0]['kwargs'])
        )(func)


class Logging_CLI_Builder(object):
    '''creates a cli function
    '''
    def __init__(self,
                 name,
                 logger,
                 check_func=None,
                 global_options_list=None):
        self.name = name
        self.logger = logger
        self.check_func = check_func
        self.global_options_list = global_options_list
        click_multi = Click_Multi(global_options_list)
        #self.multi_options = click_multi.click_multi
        dist = distribution(name).metadata
        self.version = dist['Version']
        self.author = dist['Author']
        self.email = dist['Author-email']
        self.maintainer = dist['Maintainer']
        self.maintainer_email = dist['Maintainer-email']
        self.home = dist['Home-page']
        self.summary = dist['Summary']
        self.licence = dist['License']
        self.copyright = 'Copyright (C) 2020. National Center for Genome Resources. All rights reserved.'
        self.cli_func = None

    def set_cli_func(self,cli_func):
        self.cli_func = cli_func

    def _ctx(self):
        '''private global context function'''
        return click.get_current_context

    def get_user_context_dict(self):
        '''Returns the user context dictionary.
        '''
        return self._ctx()().obj

    def init_user_context_obj(self, extra_args=None):
        '''Decorator that puts global options into context dictionary
        '''

        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                self._ctx()().obj = {}
                ctx_dict = self._ctx()().obj
                if self._ctx()().params['verbose']:
                    ctx_dict['logLevel'] = 'verbose'
                elif self._ctx()().params['quiet']:
                    ctx_dict['logLevel'] = 'quiet'
                else:
                    ctx_dict['logLevel'] = 'default'
                for key in extra_args:
                    ctx_dict[key] = self._ctx()().params[key]
                return f(*args, **kwargs)

            return wrapper

        return decorator

    def init_dual_logger(self,
                         file_log_level=DEFAULT_FILE_LOGLEVEL,
                         stderr_log_level=DEFAULT_STDERR_LOGLEVEL):
            '''Decorator to log to stderr and to logfile at different levels
            '''

            def decorator(f):
                @functools.wraps(f)
                def wrapper(*args, **kwargs):
                    # find the verbose/quiet levels from context
                    if self._ctx()().params['verbose']:
                        _log_level = logging.DEBUG
                    elif self._ctx()().params['quiet']:
                        _log_level = logging.ERROR
                    else:
                        _log_level = stderr_log_level
                    self.logger.setLevel(file_log_level)
                    stderrHandler = logging.StreamHandler(sys.stderr)
                    stderrFormatter = CleanInfoFormatter()
                    stderrHandler.setFormatter(stderrFormatter)
                    stderrHandler.setLevel(_log_level)
                    self.logger.addHandler(stderrHandler)
                    if self._ctx()().params['log']:  # start a log file
                        # If a subcommand was used, log to a file in the
                        # logs/ subdirectory of the current working directory
                        #  with the subcommand in the file name.
                        subcommand = self._ctx()().invoked_subcommand
                        if subcommand is not None:
                            logfile_name = self.name + '-' + subcommand + '.log'
                            logfile_path = Path('./logs/' + logfile_name)
                            if not logfile_path.parent.is_dir():  # create logs/ dir
                                try:
                                    logfile_path.parent.mkdir(mode=0o755, parents=True)
                                except OSError:
                                    self.logger.error(
                                        'Unable to create logfile directory "%s"',
                                        logfile_path.parent)
                                    raise OSError
                            else:
                                if logfile_path.exists():
                                    try:
                                        logfile_path.unlink()
                                    except OSError:
                                        self.logger.error(
                                            'Unable to remove existing logfile "%s"',
                                            logfile_path)
                                        raise OSError
                            logfileHandler = logging.FileHandler(str(logfile_path))
                            logfileFormatter = logging.Formatter(
                                '%(levelname)s: %(message)s')
                            logfileHandler.setFormatter(logfileFormatter)
                            logfileHandler.setLevel(file_log_level)
                            self.logger.addHandler(logfileHandler)
                    self.logger.debug('Command line: "%s"', ' '.join(sys.argv))
                    self.logger.debug('%s version %s', self.name, self.version)
                    self.logger.debug('Run started at %s', str(STARTTIME)[:-7])
                    return f(*args, **kwargs)

                return wrapper

            return decorator


    def log_elapsed_time(self):
        '''Decorator to log the elapsed time for command
        '''
        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                returnobj = f(*args, **kwargs)
                self.logger.debug(
                    'Elapsed time is %s', str(datetime.now() - STARTTIME)[:-7])
                return returnobj

            return wrapper

        return decorator

    def test_log_func(self):
        @self.cli_func.command()
        @self.log_elapsed_time()
        def test_logging():
            '''Log at different severity levels.
            '''
            self.logger.debug('debug message')
            self.logger.info('info message')
            self.logger.warning('warning message')
            self.logger.error('error message')

    def show_context_func(self):
        @self.cli_func.command()
        def show_context_dict():
            '''Print the global context dictionary.
            '''
            user_ctx = self.get_user_context_dict()
            self.logger.info('User context dictionary:')
            for key in user_ctx.keys():
                self.logger.info('   %s: %s', key, user_ctx[key])


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bionorm']

package_data = \
{'': ['*'], 'bionorm': ['templates/*']}

install_requires = \
['biopython>=1.76,<2.0',
 'click>=7.0,<8.0',
 'click_plugins>=1.1.1,<2.0.0',
 'coverage>=5.0.3,<6.0.0',
 'importlib_metadata>=1.5.0,<2.0.0',
 'ruamel.yaml>=0.16.6,<0.17.0',
 'sequencetools>=0.0.5,<0.0.6']

entry_points = \
{'console_scripts': ['bionorm = bionorm:cli']}

setup_kwargs = {
    'name': 'bionorm',
    'version': '0.1.1',
    'description': 'normalize and verify genomic data',
    'long_description': 'bionorm\n======\ncreating, searching, and analyzing phylogenetic signatures from genomes or reads of DNA.\n\nPrerequisites\n-------------\nPython 3.6 or greater is required.\n\nRun-time dependencies of bionorm are : ::\n\n    biopython, click, ruamel.yaml, and sequencetools.\n\nbionorm uses the poetry dependency manager\n\nInstallation\n------------\nThis package is tested under Linux and MacOS using Python 3.7.\ninstall via pip (or pip3 under some distributions) : ::\n\n     pip install bionorm\n\nIf you wish to develop bionorm,  download a `release <https://github.com/ncgr/bionorm/releases>`_\nand in the top-level directory: ::\n\n\tpip install --editable .\n\nIf you wish to have pip install directly from git, use this command: ::\n\n\tpip install git+https://github.com/ncgr/bionorm.git\n\n\nUsage\n-----\nInstallation puts a single script called ``bionorm`` in your path.  The usage format is::\n\n    bionorm [GLOBALOPTIONS] COMMAND [COMMANDOPTIONS] [ARGS]\n\nA listing of commands is available via ``bionorm --help``.  Current available commands are:\n\n============================= ====================================================\n  busco                       Perform BUSCO checks.\n  detector                    Detect/correct incongruencies among files.\n  fasta                       Check for GFF/FASTA consistency.\n  generate_readme             Generates a README file with details of genome.\n  index                       Indexes FASTA file.\n\n============================= ====================================================\n\n\n+-------------------+------------+------------+\n| Latest Release    | |pypi|     | |bionorm|  |\n+-------------------+------------+            +\n| GitHub            | |repo|     |            |\n+-------------------+------------+            +\n| License           | |license|  |            |\n+-------------------+------------+            +\n| Documentation     | |rtd|      |            |\n+-------------------+------------+            +\n| Travis Build      | |travis|   |            |\n+-------------------+------------+            +\n| Coverage          | |coverage| |            |\n+-------------------+------------+            +\n| Code Grade        | |codacy|   |            |\n+-------------------+------------+            +\n| Dependencies      | |depend|   |            |\n+-------------------+------------+            +\n| Issues            | |issues|   |            |\n+-------------------+------------+------------+\n\n\n.. |bionorm| image:: docs/normal.jpg\n     :alt: Make me NORMAL, please!\n\n.. |pypi| image:: https://img.shields.io/pypi/v/bionorm.svg\n    :target: https://pypi.python.org/pypi/bionorm\n    :alt: Python package\n\n.. |repo| image:: https://img.shields.io/github/commits-since/ncgr/bionorm/0.1.svg\n    :target: https://github.com/ncgr/bionorm\n    :alt: GitHub repository\n\n.. |license| image:: https://img.shields.io/badge/License-BSD%203--Clause-blue.svg\n    :target: https://github.com/ncgr/bionorm/blob/master/LICENSE.txt\n    :alt: License terms\n\n.. |rtd| image:: https://readthedocs.org/projects/bionorm/badge/?version=latest\n    :target: http://bionorm.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Server\n\n.. |travis| image:: https://img.shields.io/travis/ncgr/bionorm.svg\n    :target:  https://travis-ci.org/ncgr/bionorm\n    :alt: Travis CI\n\n.. |codacy| image:: https://api.codacy.com/project/badge/Grade/b23fc0c167fc4660bb649320e14dac7f\n    :target: https://www.codacy.com/gh/ncgr/bionorm?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=ncgr/bionorm&amp;utm_campaign=Badge_Grade\n    :alt: Codacy.io grade\n\n.. |coverage| image:: https://codecov.io/gh/ncgr/bionorm/branch/master/graph/badge.svg\n    :target: https://codecov.io/gh/ncgr/bionorm\n    :alt: Codecov.io test coverage\n\n.. |issues| image:: https://img.shields.io/github/issues/ncgr/bionorm.svg\n    :target:  https://github.com/ncgr/bionorm/issues\n    :alt: Issues reported\n\n.. |depend| image:: https://api.dependabot.com/badges/status?host=github&repo=ncgr/bionorm\n     :target:https://app.dependabot.com/accounts/ncgr/repos/236847525\n     :alt: dependabot dependencies\n\n',
    'author': 'Connor Cameron',
    'author_email': 'ctc@ncgr.org',
    'maintainer': 'Joel Berendzen',
    'maintainer_email': 'joelb@ncgr.org',
    'url': 'https://github.com/ncgr/ncgr',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python

import os
import sys


def cli():
    '''Normalizer -- Normalize Files to Data Store Standards

           The normalizer will detect the input file type.
           Currently recognizes case independent fasta, fna, fa, gff3, gff
           START WITH THE PREFIX COMMAND UNLESS YOU JUST WANT TO INDEX!
               This stages the canonical directory structure

       USAGE: normalizer <mode> [options]

       Current Modes:

           prefix        (will sort GFF3 Files)
           index         (Canonical indexing for datastore)
           readme        (Generates README files for directories)
           extract_fasta (extracts mRNA, CDS and Peptides from GFF3)
           busco         (Runs BUSCO on the target)
           checksums     (Generate md5s for all files in target dir)

       Please run `normalizer <TOOL> --help` for individual usage
    '''
    if not len(sys.argv) > 1:
        print(cli.__doc__)
        sys.exit(1)
    mode = sys.argv[1].lower()  # get mode
    sys.argv = [mode] + sys.argv[2:]  # makes usage for application correct
    if mode == '--help' or mode == '-h':
        print(cli.__doc__)
        sys.exit(1)
    if mode == 'index':  # determine file type and index
        from . import index
        index.cli()
    if mode == 'prefix':  # determine file type and prefix, sorts gff3
        from . import prefix
        prefix.cli()
    if mode == 'readme':  # generate readmes from template and target
        from . import generate_readme
        generate_readme.cli()
    if mode == 'extract_fasta':  # Extract FASTA features from GFF3
        from . import extract_fasta
        extract_fasta.cli()
    if mode == 'busco':  # Run BUSCO with mode and lineage on target
        from . import busco_normalizer
        busco_normalizer.cli()


if __name__ == '__main__':
    cli()



#!/usr/bin/env python

from .helper import check_file, return_filehandle
import os
import sys
import logging
import re
import subprocess
import hashlib
import requests
import json
from glob import glob
from os import path
sys.path.append(path.dirname(path.dirname(path.abspath(__file__))))


class genome_main:

    def __init__(self, detector, **kwargs):
        self.detector = detector
        self.target = detector.target
        self.logger = detector.logger

    def run(self):
        '''Runs checks'''
        logger = self.logger
        target = self.target
        logger.info('Performing Naming Checks for {}\n'.format(target))
        if not self.check_genome_main():
            logger.error('Naming Check FAILED')
            return False
        logger.info('Naming Looks Correct\n')
        logger.info('Checking Genome FASTA\n')
        if not self.check_genome_fasta():
            logger.error('FASTA Check FAILED')
            return False
        logger.info('ALL GENOME CHECKS PASSED\n')
        return True

    def check_genome_main(self):
        '''accepts a list of genome attributes split by "."

           https://github.com/LegumeFederation/datastore/issues/23

           checks these file attributes to ensure they are correct
        '''
        target = self.target
        logger = self.logger
        attr = os.path.basename(target).split('.')  # split on delimiter
        if len(attr) != 7:  # should be 7 fields
            logger.error('File did not have 7 fields! {}'.format(attr))
            return False
        if len(attr[0]) != 5:  # should be 5 letter prefix
            logger.error('File must have 5 letter prefix, not {}'.format(
                attr[0]))
            return False
        if not attr[2].startswith('gnm'):  # should be gnm type
            logger.error('File should have gnm in field 3, not {}'.format(
                attr[2]))
            return False
        gnm_v = attr[2].replace('gnm', '')
        try:
            int(gnm_v)
        except ValueError:  # best way to check for int in python2
            logger.error('gnm version must be integer not {}'.format(gnm_v))
            return False
        if (len(gnm_v.split('.')) > 1):  # check for float
            logger.error('gnm version must be integer not {}'.format(gnm_v))
            return False
        if not attr[5] == 'fna':  # should be fna type
            logger.error('File should be fna not {}'.format(attr[5]))
            return False
        if not attr[6] == 'gz':  # should be gzip compressed
            logger.error('Last field should be gz, not {}'.format(attr[6]))
            return False
        logger.info('Genome Naming Looks Correct')
        return True

    def check_genome_fasta(self):
        '''Confirms that headers in fasta genome_main conform with standard

           PUT SOME RULE REFERENCE HERE
        '''
        logger = self.logger
        fasta = self.target  # get fasta file
        attr = os.path.basename(fasta).split('.')  # get attributes for naming
        true_header = '.'.join(attr[:3])
        fh = return_filehandle(fasta)  # get file handle, text/gz
        # grab header and description
        re_header = re.compile(r"^>(\S+)\s*(.*)")
        passed = True
        with fh as gopen:
            for line in gopen:
                line = line.rstrip()
                if not line:
                    continue
                if re_header.match(line):  # check for fasta header
                    hid = re_header.search(line)
                    if hid:
                        logger.debug(hid.groups(0))
                        if isinstance(hid, str):  # check for tuple
                            hid = hid.groups(0)
                        else:
                            hid = hid.groups(0)[0]  # get id portion of header
                    else:
                        logger.error('Header {} looks odd...'.format(line))
                        return False
                    logger.debug(hid)
                    self.detector.fasta_ids[hid] = 1
                    standard_header = true_header + '.' + hid
                    if not hid.startswith(true_header):
                        logger.warning(('Inconsistency {} '.format(hid) +
                                        'Should be {}'.format(standard_header))
                                       )
                        passed = False
        return passed


class gene_models_main:

    def __init__(self, detector, **kwargs):
        self.detector = detector
        self.target = detector.target
        self.logger = detector.logger
        self.fasta_ids = detector.fasta_ids

    def run(self):
        '''Run checks'''
        logger = self.logger
        target = self.target
        logger.info('Checking Gene Models File Naming...')
        if not self.check_gene_models_main():
            logger.error('Gene Model File Naming FAILED')
            return False
        logger.info('Naming Looks Correct\n')
        logger.info('Validating GFF3 with gt...')
        if not self.gt_gff3_validate():
            logger.error('gt GFF3 Validation FAILED')
            return False
        logger.info('GFF3 is Valid\n')
        logger.info('Checking Congruency Between Genome and Gene Models...')
        if not self.check_seqid_attributes():
            logger.error('Genome and Gene Models are not Congruent FAILED')
            return False
        logger.info('ALL GENE MODELS CHECKS PASSED\n')
        return True

    def check_gene_models_main(self):
        '''accepts a list of annotation attributes split by "."

           https://github.com/LegumeFederation/datastore/issues/23

           checks these file attributes to ensure they are correct
        '''
        target = self.target
        logger = self.logger
        attr = os.path.basename(target).split('.')  # split on delimiter
        if len(attr) != 8:  # should be 8 fields
            logger.error('File did not have 7 fields! {}'.format(attr))
            return False
        if len(attr[0]) != 5:  # should be 5 letter prefix
            logger.error('File must have 5 letter prefix, not {}'.format(
                attr[0]))
            return False
        if not attr[2].startswith('gnm'):  # should be gnm type
            logger.error('File should have gnm in field 3, not {}'.format(
                attr[2]))
            return False
        gnm_v = attr[2].replace('gnm', '')
        try:
            int(gnm_v)
        except ValueError:  # best way to check for int in python2
            logger.error('gnm version must be integer not {}'.format(gnm_v))
            return False
        if (len(gnm_v.split('.')) > 1):  # check for float
            logger.error('gnm version must be integer not {}'.format(gnm_v))
            return False
        if not attr[3].startswith('ann'):  # should be gnm type
            logger.error('File should have ann in field 4, not {}'.format(
                attr[2]))
            return False
        ann_v = attr[3].replace('ann', '')
        try:
            int(ann_v)
        except ValueError:  # best way to check for int in python2
            logger.error('ann version must be integer not {}'.format(ann_v))
            return False
        if (len(ann_v.split('.')) > 1):  # check for float
            logger.error('ann version must be integer not {}'.format(gnm_v))
            return False
        if not attr[6] == 'gff3':  # should be gff3 type
            logger.error('File should be gff3 not {}'.format(attr[6]))
            return False
        if not attr[7] == 'gz':  # should be gzip compressed
            logger.error('Last field should be gz, not {}'.format(attr[6]))
            return False
        return True

    def gt_gff3_validate(self):
        '''Confirms that gff3 files pass gt validation

           https://github.com/LegumeFederation/datastore/issues/23
        '''
        gff = self.target
        logger = self.logger
        gff_name = os.path.basename(gff)
        gt_report = './{}_gt_gff3validator_report.txt'.format(gff_name)
        gt_cmd = '(gt gff3validator {} 2>&1) > {}'.format(gff,
                                                          gt_report)
        logger.debug(gt_cmd)
        exit_val = subprocess.call(gt_cmd, shell=True)  # get gt exit_val
        logger.debug(exit_val)
        if exit_val:
            return False
        return True

    def check_seqid_attributes(self):
        '''Confirms that gff3 seqid exists in genome_main if provided

           checks ID and Name from gff3 attributes field

           https://github.com/LegumeFederation/datastore/issues/23
        '''
        gff = self.target
        logger = self.logger
        fasta_ids = self.fasta_ids  # list of FASTA IDS from Reference
        logger.debug(fasta_ids)
        fh = return_filehandle(gff)
        file_name = os.path.basename(gff)
        # ID should start with this string
        true_id = '.'.join(file_name.split('.')[:4])
        true_name = file_name.split('.')[0]  # maybe this should include infra
#        get_id_name = re.compile("^ID=(.+?);.*Name=(.+?);")
        get_id = re.compile('ID=([^;]+)')
        lines = 0
        seen = {}
        passed = True
        with fh as gopen:
            for line in gopen:
                line = line.rstrip()
                lines += 1
                if not line or line.startswith('#'):
                    continue
                columns = line.split('\t')  # get gff3 fields
                seqid = columns[0]  # seqid according to the spec
                seqid = seqid.rstrip()
                logger.debug(line)
                logger.debug(seqid)
                if self.fasta_ids:  # if genome_main make sure seqids exist
                    if seqid not in fasta_ids:  # fasta header check
                        logger.debug(seqid)
                        if seqid not in seen:
                            logger.error('{} not found in genome_main'.format(
                                seqid))
                            seen[seqid] = 1
                        passed = False
                feature_type = columns[2]  # get type
                attributes = columns[8]  # attributes ';' delimited
                logger.debug(feature_type)
                if feature_type != 'gene':  # only check genes (for now)
                    continue
#                if not get_id_name.match(attributes):  # check for ID and Name
                if not get_id.search(attributes):  # check for ID and Name
                    logger.error('No ID and Name attributes. line {}'.format(
                        lines))
                    passed = False
                else:
                    #                    groups = get_id_name.search(attributes).groups()
                    feature_id = get_id.search(attributes).group(1)
                    logger.debug(feature_id)
#                    if len(groups) != 2:  # should just be ID and Name
#                        logger.error('too many groups detected: {}'.format(
#                                                                      groups))
#                        passed = False
#                    (feature_id, feature_name) = groups
                    if not feature_id.startswith(true_id):  # check id
                        logger.error('gene feature id, should start with ' +
                                     '{} line {}'.format(true_id, lines))
                        passed = False
#                    if not feature_name.startswith(true_name):
#                        logger.error('feature name, should start with ' +
#                                     '{} line {}'.format(true_name, lines))
#                        passed = False
        return passed


class protein:

    def __init__(self, detector, **kwargs):
        self.detector = detector
        self.target = detector.target
        self.logger = detector.logger

    def run(self):
        '''Run Checks for protein'''
        logger = self.logger
        logger.info('protein check {}'.format(self.target))
        if not self.check_protein():
            logger.error('protein file naming checks FAILED')
            return False
        logger.info('ALL protein CHCEKS PASSED')
        return True

    def check_protein(self):
        '''accepts a list of annotation attributes split by "."

           https://github.com/LegumeFederation/datastore/issues/23

           checks these file attributes to ensure they are correct
        '''
        target = self.target
        logger = self.logger
        attr = os.path.basename(target).split('.')  # split on delimiter
        if len(attr) != 8:  # should be 8 fields
            logger.error('File did not have 7 fields! {}'.format(attr))
            return False
        if len(attr[0]) != 5:  # should be 5 letter prefix
            logger.error('File must have 5 letter prefix, not {}'.format(
                attr[0]))
            return False
        if not attr[2].startswith('gnm'):  # should be gnm type
            logger.error('File should have gnm in field 3, not {}'.format(
                attr[2]))
            return False
        gnm_v = attr[2].replace('gnm', '')
        try:
            int(gnm_v)
        except ValueError:  # best way to check for int in python2
            logger.error('gnm version must be integer not {}'.format(gnm_v))
            return False
        if (len(gnm_v.split('.')) > 1):  # check for float
            logger.error('gnm version must be integer not {}'.format(gnm_v))
            return False
        if not attr[3].startswith('ann'):  # should be gnm type
            logger.error('File should have ann in field 4, not {}'.format(
                attr[2]))
            return False
        ann_v = attr[3].replace('ann', '')
        try:
            int(ann_v)
        except ValueError:  # best way to check for int in python2
            logger.error('ann version must be integer not {}'.format(ann_v))
            return False
        if (len(ann_v.split('.')) > 1):  # check for float
            logger.error('ann version must be integer not {}'.format(gnm_v))
            return False
        if not attr[6] == 'faa':  # should be gff3 type
            logger.error('File should be gff3 not {}'.format(attr[6]))
            return False
        if not attr[7] == 'gz':  # should be gzip compressed
            logger.error('Last field should be gz, not {}'.format(attr[6]))
            return False
        return True


class protein_primaryTranscript:

    def __init__(self, detector, **kwargs):
        self.detector = detector
        self.target = detector.target
        self.logger = detector.logger

    def run(self):
        '''Run Checks for protein_primaryTranscript'''
        logger = self.logger
        logger.info('protein_primaryTranscript check {}'.format(self.target))
        if not self.check_protein_primaryTranscript():
            logger.error('protein_primaryTranscript naming checks FAILED')
            return False
        logger.info('ALL protein_primaryTranscript CHCEKS PASSED')
        return True

    def check_protein_primaryTranscript(self):
        '''accepts a list of annotation attributes split by "."

           https://github.com/LegumeFederation/datastore/issues/23

           checks these file attributes to ensure they are correct
        '''
        target = self.target
        logger = self.logger
        attr = os.path.basename(target).split('.')  # split on delimiter
        if len(attr) != 8:  # should be 8 fields
            logger.error('File did not have 7 fields! {}'.format(attr))
            return False
        if len(attr[0]) != 5:  # should be 5 letter prefix
            logger.error('File must have 5 letter prefix, not {}'.format(
                attr[0]))
            return False
        if not attr[2].startswith('gnm'):  # should be gnm type
            logger.error('File should have gnm in field 3, not {}'.format(
                attr[2]))
            return False
        gnm_v = attr[2].replace('gnm', '')
        try:
            int(gnm_v)
        except ValueError:  # best way to check for int in python2
            logger.error('gnm version must be integer not {}'.format(gnm_v))
            return False
        if (len(gnm_v.split('.')) > 1):  # check for float
            logger.error('gnm version must be integer not {}'.format(gnm_v))
            return False
        if not attr[3].startswith('ann'):  # should be gnm type
            logger.error('File should have ann in field 4, not {}'.format(
                attr[2]))
            return False
        ann_v = attr[3].replace('ann', '')
        try:
            int(ann_v)
        except ValueError:  # best way to check for int in python2
            logger.error('ann version must be integer not {}'.format(ann_v))
            return False
        if (len(ann_v.split('.')) > 1):  # check for float
            logger.error('ann version must be integer not {}'.format(gnm_v))
            return False
        if not attr[6] == 'faa':  # should be gff3 type
            logger.error('File should be gff3 not {}'.format(attr[6]))
            return False
        if not attr[7] == 'gz':  # should be gzip compressed
            logger.error('Last field should be gz, not {}'.format(attr[6]))
            return False
        return True


class readme_md:  # need to populate this correctly later

    def __init__(self, detector, **kwargs):
        self.detector = detector
        self.target = detector.target
        self.logger = detector.logger
        self.fasta_ids = detector.fasta_ids

    def validate_checksum(self, md5_file, check_me):
        '''Get md5 checksum for file and compare to expected'''
        logger = self.logger
        fh = return_filehandle(md5_file)
        hash_md5 = hashlib.md5()
        check_sum_target = ''
        switch = 0
        with fh as copen:
            for line in copen:
                line = line.rstrip()
                if not line or line.startswith('#'):
                    continue
                fields = line.split()
                check_sum = fields[0]
                filename = fields[1]
                logger.debug('check_sum: {}, filename: {}'.format(
                    check_sum,
                    filename))
                if not check_sum and filename:
                    logger.error('Could not find sum and name for {}'.format(
                        line))
                if filename == os.path.basename(check_me):
                    logger.info('Checksum found for {}'.format(filename))
                    check_sum_target = check_sum
                    switch = 1
        if not switch:
            logger.error('Could not find checksum for {}'.format(check_me))
            sys.exit(1)
        with open(check_me, "rb") as copen:
            for chunk in iter(lambda: copen.read(4096), b""):  # 4096 buffer
                hash_md5.update(chunk)
        target_sum = hash_md5.hexdigest()  # get sum
        logger.debug(target_sum)
        logger.debug(check_sum_target)
        if target_sum != check_sum_target:  # compare sums
            logger.error(('Checksum for file {} {} '.format(check_me,
                                                            target_sum) +
                          'did not match {}'.format(check_sum_target)))
            sys.exit(1)
        logger.info('Checksums checked out, moving on...')

    def validate_doi(self, readme):
        '''Parse README.<key>.yml and get publication or dataset DOIs

           Uses http://www.doi.org/factsheets/DOIProxy.html#rest-api
        '''
        logger = self.logger
        publication_doi = ''
        dataset_doi = ''
        object_dois = {}
